/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/ 
/*
  Copyright (C) 1998-1999 Johan Redestig
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/****************************************************************************
12/19/00 - Bobby G. Vinyard (Message)
  - Added !PopupRecycleBin and !PopupRecentDocuments (this works correctly
    under W2K where as !DynamicFolder:$recent$ display everything
11/27/00 - Joachim Calvert (NeXTer)
  - Fixed the RevID so that the right version number is reported
11/15/00 - Kevin Schaffer (Maduin)
  - added PopupDefaultIcon, PopupIconSpacing, and PopupIconSize
  - added PopupTitleAlignment and PopupEntryAlignment
  - fixed a resource leak - MenuItem::~MenuItem() used DeleteObject to destroy an icon
11/12/00 - Bobby G. Vinyard (Message)
  - Removed 2000-10-18 additions to popup2 for right click open
  - Implemented right click context menus for the following items: !PopupFolder,
    !PopupDynamicFolder, !DynamicFolder, !PopupMyComputer, !PopupNetwork, !PopupControlPanel,
    !PopupPrinters (this code is by no means perfect, i.e. sub menus on the context menu
    do not function, i.e. the send to menu)
10/15/00 - Bobby G. Vinyard (Message)
  - Right clicking on a menu item that is a !PopupFolder or !PopupDynamicFolder, or one of
    of their subfolders. will cause the directory to be opened in a single pane explorer
    window
09/21/00 - Kevin Schaffer (Maduin)
  - Added .icon support
  - Added NoPopupTitleText
  - PopupBottomPix can now be a different height than PopupTitlePix
09/18/00 - Kevin Schaffer (Maduin)
  - Added support for some useful shell folders
    - !PopupMyComputer
	- !PopupNetwork
	- !PopupControlPanel
	- !PopupPrinters
08/31/00 - Joachim Calvert (NeXTer)
  - Altered the WM_POPUP message to pass both mouse coords in lParam
08/28/00 - Joachim Calvert (NeXTer)
  - Now handles scrolling of long menus, courtesy of Sylvain Rouquette (Syl)
  - New setting: PopupScrollSpeed <int>, default is 1
  - Errata: PopupTitleBold only works in conjunction with PopupFontFace, due
    to the method for retreiving the default font
08/27/00 - Joachim Calvert (NeXTer)
  - Should now handle gradients gracefully on newer builds, let me know if any
    problems occur
  - Added the settings PopupTitleBold <bool> and PopupEntryFontHeight <int>
  - Made !DynamicFolder menus display the caption of their parent
  - Fixed the vertical alignment of the text, should look better now
08/24/00 - Joachim Calvert (NeXTer)
  - Changed the min width default to 128px
08/17/00 - Joachim Calvert (NeXTer)
  - Hidden files and folders are no longer shown
  - Hides the extensions for .bat and .pif files
08/15/00 - Bobby G. Vinyard (Message)
  - Fixed a bug in !PopupDynamicFolder not showing
08/05/00 - Joachim Calvert (NeXTer)
  - Made the folder arrow static (doesn't resize)
  - Now takes the folder arrow into account when calculating the width of items
07/21/00 - Bobby G. Vinyard (Message)
  - Added !PopupDynamicFolder for compatiability with old Popup
  - Added folder merging capability to !DynamicFolder, !PopupDynamicFolder
    and !PopupFolder
****************************************************************************/
#include "stdafx.h"
#include "PopupMaker.h"
#include <winuser.h> 
// menu items
#include "MenuItem.h"
#include "FolderItem.h"
#include "PopupMenu.h"
#include "SeparatorItem.h"
#include "CommandItem.h"
#include "TimeItem.h"
#include "HeaderItem.h"
#include "DirectoryFolder.h"

#ifndef _MINI
	#include "RemoteAmpItem.h"
	#include "RunItem.h"
	#include "../lsapi/safestr.h"
#endif /* _MINI */

#include "TaskFolder.h"
#include "TaskItem.h"
#include "DesktopMenu.h"
#include "TaskMenu.h"
#include "ShellFolder.h"
#include "Re5ources.h"

// Painters
#ifndef _MINI
	#include "BmpPainter.h"
	#include "GradientPainter.h"
#endif /* _MINI */

#include "SolidPainter.h"

#include "../lsapi/lsapi.h"

#include <io.h>
#ifndef _MINI
	#include <shlwapi.h>
#endif /* _MINI */
#include <vector>
#include <algorithm>

extern vector<PopupMenu*> g_PopupMenues;

const char rcsRevision[] = "$Revision: 1.11+EVWM $"; // Our Version
const char rcsId[] = "$Popup2 with EVWM functionality: v1.11 + EVWM functionality by normalperson $"; // The Full RCS ID.

//int test2 = 1;
//int PopupMaker::m_snSubmenuHeight = 16;
//Painter * PopupMaker::m_spEntry = NULL;
//Painter * PopupMaker::m_spSelEntry = NULL;
//BOOL CALLBACK EnumWindowsProc(HWND hWnd, LPARAM lParam);

//Painter * g_pSelEntry;
//Painter * g_pEntry;
//int g_nSubmenuHeight;

extern PopupMaker *pPopupMaker;

/*IgnoreList *ignoreTag;

BOOL PopupMaker::EnumWindowsProc(HWND hWnd, LPARAM lParam)
{
	if (IsAppWindowWithFilter(hWnd, pPopupMaker->m_nTaskFilterLevel))
	{
		if ( pPopupMaker->numIgnore > 0 )
		{
			char title[1024];
			GetWindowText(hWnd, title, sizeof(title) );
			
			for (int i=0; i<pPopupMaker->numIgnore; ++i)
			{
				if (strstr(title, ignoreTag[i].match))
					return TRUE;
			}
		}
		char * m_pszTitle;
		char pszWindowText[1024];
		if ( 0 != GetWindowText(hWnd, pszWindowText, sizeof(pszWindowText)) )
			m_pszTitle = pszWindowText != NULL ? strdup(pszWindowText) : NULL ;
		else
			m_pszTitle = " NULL ";

		PopupMenu *menu = (PopupMenu *) lParam;
		TaskItem *item = new TaskItem(hWnd);
	
		item->SetActivePainter(pPopupMaker->m_pSelEntry);
		item->SetPainter(pPopupMaker->m_pEntry);
		item->SetHeight(pPopupMaker->m_nSubmenuHeight);

		if (pPopupMaker->m_bTaskIcons)
			item->SetIcon(GetIconFromWindowWithTimeout(hWnd, FALSE,pPopupMaker->m_nGetIconTimeout));
			menu->AddMenuItem(item);
	}
	return TRUE;
}*/

PopupMaker::PopupMaker()
{
	m_pRoot = NULL;
	m_pBackground = NULL;
	m_pSeparator = NULL;
	m_pBottom = NULL;
}

PopupMaker::~PopupMaker()
{
	// unregister LS messages
	if (m_pRoot)
	{
		int msgs[] = {
		                 LM_POPUP, LM_HIDEPOPUP, LM_GETREVID, 0
		             };

		SendMessage(GetLitestepWnd(), LM_UNREGISTERMESSAGE, (WPARAM) m_pRoot->GetWindow(), (LPARAM) msgs);
	}

	DeleteObject(m_hfnTitle);
	DeleteObject(m_hfnEntry);
	DeleteObject(m_hfnFolder);

	delete m_pBackground;
	delete m_pTitle;
	delete m_pEntry;
	delete m_pSelEntry;
	delete m_pSeparator;
	delete m_pBottom;

	if (m_hbmArrow)
	{
		DeleteObject(m_hbmArrow);
		DeleteObject(m_hbmSelArrow);
	}

	// delete all PopupMenu object, this is important not just only for
	// not leaking memory but (which is important since one recycles
	// litestep from time to time. litestep will crash upon recycle otherwise
	while (!g_PopupMenues.empty())
		delete g_PopupMenues[0];

	// delete the painters that we have allocated.
	for (int i = 0; i < m_Painters.size(); ++i)
		delete m_Painters[i];
	
/*	for (i=0;i<numIgnore;++i)
		delete [] ignoreTag[i].match;
	if (ignoreTag)
		free(ignoreTag); */
}

void PopupMaker::Show(int index)
{
	m_RootMenus[index]->Show();
	SetFocus(m_RootMenus[index]->GetWindow());
}

void PopupMaker::Hide(int index)
{
	for (int i = 0; i < m_RootMenus.size(); ++i)
	{
//		m_RootMenus[index]->DeleteMenuItems();
		m_RootMenus[i]->Hide(HIDE_PARENTS);
		m_RootMenus[i]->Hide(HIDE_CHILDREN);
	}
}

void PopupMaker::ShowDesktopMenu()
{
	m_pDesktopMenu->Show();
	SetFocus(m_pDesktopMenu->GetWindow());
}

void PopupMaker::ShowTaskMenu()
{
	m_pTaskMenu->Show();
	SetFocus(m_pTaskMenu->GetWindow());
}

void PopupMaker::Initialize(HWND hParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	hInst = dllInst;

	// we are managing the system errors our seleves. this will prevent
	// the system to show a error message box if we would try to l
	UINT nErrorMode = SetErrorMode(1);
	m_bTransparent = GetRCBool("PopupNotTransparent", FALSE); // draw transparent by default
	m_bShowGradients = GetRCBoolDef("PopupPaintGradient", TRUE);

	m_bHeader = GetRCBool("PopupNoTitles", FALSE);
	if (m_bHeader)
		m_bActiveHeaders = GetRCBoolDef("PopupActiveTitles", FALSE);
	else 
		m_bActiveHeaders = TRUE;
	m_bHeaderText = GetRCBool("PopupNoTitleText", FALSE);


	// Set the painters ...
	GetRCString("PopupTitlePix", m_pszTitlePix, "", MAX_RC_STRING);
	GetRCString("PopupEntryPix", m_pszEntryPix, "", MAX_RC_STRING);
	GetRCString("PopupSelEntryPix", m_pszSelEntryPix, "", MAX_RC_STRING);
	GetRCString("PopupFolderPix", m_pszFolderPix, m_pszEntryPix, MAX_RC_STRING);
	if (stricmp(m_pszFolderPix, m_pszEntryPix) == 0)
	{
		DirectoryFolder::SetPaintFolderIcon(TRUE);
	}
	else
	{
		DirectoryFolder::SetPaintFolderIcon(FALSE);
	}
	GetRCString("PopupSelFolderPix", m_pszSelFolderPix, m_pszSelEntryPix, MAX_RC_STRING);
	GetRCString("PopupSeparatorPix", m_pszSeparatorPix, "", MAX_RC_STRING);
	GetRCString("PopupBottomPix", m_pszBottomPix, "", MAX_RC_STRING);

	// Base colors
	m_nTitleColor = GetRCColor("PopupTitleBgColor", GetSysColor(COLOR_ACTIVECAPTION));
	m_nInActiveTitleColor = GetRCColor("PopupInActiveTitleBgColor", GetSysColor(COLOR_INACTIVECAPTION));
	m_nEntryColor = GetRCColor("PopupEntryBgColor", GetSysColor(COLOR_BTNFACE));
	m_nSelEntryColor = GetRCColor("PopupSelEntryBgColor", GetSysColor(COLOR_HIGHLIGHT));
	m_nFolderColor = GetRCColor("PopupFolderBgColor", m_nEntryColor);
	m_nSelFolderColor = GetRCColor("PopupSelFolderBgColor", m_nSelEntryColor);

#ifndef _MINI
	// Gradient colors
	m_nGradientTitle = GetRCColor("PopupGradientTitle", m_nTitleColor);
	m_bGradientTitle = m_nGradientTitle != m_nTitleColor;
	m_nGradientEntry = GetRCColor("PopupGradientEntry", m_nEntryColor);
	m_bGradientEntry = m_nGradientEntry != m_nEntryColor;
	m_nGradientFolder = GetRCColor("PopupGradientFolder", m_nFolderColor);
	m_bGradientFolder = m_nGradientFolder != m_nFolderColor;
#endif /* _MINI */
	// Text colors
	m_nTitleTextColor = GetRCColor("PopupTitleColor", GetSysColor(COLOR_CAPTIONTEXT));
	m_nEntryTextColor = GetRCColor("PopupEntryColor", GetSysColor(COLOR_MENUTEXT));
	m_nSelEntryTextColor = GetRCColor("PopupSelEntryColor", GetSysColor(COLOR_HIGHLIGHTTEXT));
	m_nFolderTextColor = GetRCColor("PopupFolderColor", m_nEntryTextColor);
	m_nSelFolderTextColor = GetRCColor("PopupSelFolderColor", m_nSelEntryTextColor);

	// 3d decoration colors
	m_nBevelLightColor = GetRCColor("PopupBevelLightColor", GetSysColor(COLOR_3DHILIGHT));
	m_nBevelDarkColor = GetRCColor("PopupBevelDarkColor", GetSysColor(COLOR_3DSHADOW));

	m_pTitle = MakePainter(m_pszTitlePix, m_nTitleColor, m_nGradientTitle, m_bGradientTitle, FALSE);
	m_pEntry = MakePainter(m_pszEntryPix, m_nEntryColor, m_nGradientEntry, m_bGradientEntry, FALSE);
	m_pSelEntry = MakePainter(m_pszSelEntryPix, m_nSelEntryColor, m_nGradientEntry, m_bGradientEntry, TRUE);
	m_pFolder = MakePainter(m_pszFolderPix, m_nFolderColor, m_nGradientFolder, m_bGradientFolder, FALSE);
	m_pSelFolder = MakePainter(m_pszSelFolderPix, m_nSelFolderColor, m_nGradientFolder, m_bGradientFolder, TRUE);

#ifndef _MINI
	if (lstrlen(m_pszSeparatorPix) > 0)
	{
		BITMAP bm;
		HBITMAP hbmSeparator = LoadLSImage(m_pszSeparatorPix, NULL);
		GetObject(hbmSeparator, sizeof(BITMAP), &bm);
		m_pSeparator = new BmpPainter(hbmSeparator, m_bTransparent);
		m_nSeparatorHeight = bm.bmHeight;
	}

	if (lstrlen(m_pszBottomPix) > 0)
	{
		BITMAP bm;
		HBITMAP hbmBottom = LoadLSImage(m_pszBottomPix, NULL);
		GetObject(hbmBottom, sizeof(BITMAP), &bm);
		m_pBottom = new BmpPainter(hbmBottom, m_bTransparent);
		m_nBottomHeight = bm.bmHeight;
	} 

	char szArrowImage[MAX_PATH];
	char szSelArrowImage[MAX_PATH];

	GetRCString("PopupArrowImage", szArrowImage, "", MAX_PATH);
	GetRCString("PopupSelArrowImage", szSelArrowImage, szArrowImage, MAX_PATH);

	if (szArrowImage[0])
	{
		m_hbmArrow = LoadLSImage(szArrowImage, NULL);
		m_hbmSelArrow = LoadLSImage(szSelArrowImage, NULL);

		MenuItem::SetArrow(m_hbmArrow);
		MenuItem::SetSelArrow(m_hbmSelArrow);
	}
	else 
#endif /* _MINI */
	{
		m_hbmArrow = NULL;
		m_hbmSelArrow = NULL;
	}


	GetRCString("PopupDefaultIcon", m_pszDefaultIcon, "", MAX_RC_STRING);
	m_bIcons = GetRCBool("PopupIcons", TRUE);
	m_bTaskIcons = GetRCBoolDef("PopupTaskIcons", TRUE);
	m_bShellIcons = GetRCBoolDef("PopupShellIcons", TRUE);
	m_bPathIcons = GetRCBoolDef("PopupPathIcons", TRUE);
	m_nIconSpacing = GetRCInt("PopupIconSpacing", 4);
	m_nIconSize = GetRCInt("PopupIconSize", 16);
	m_bMThreadMScroll = GetRCBoolDef("PopupDebugMultiThreadMouseScroll",FALSE);

	// I'm impatient when it comes to popup speed!  >:(  -- (normalperson)
	m_nGetIconTimeout = GetRCInt("PopupGetIconTimeout",1);

	MenuItem::SetIconSpacing(m_nIconSpacing);
	MenuItem::SetIconSize(m_nIconSize);
	MenuItem::SetIconsOnRight(GetRCBool("PopupIconsOnRight", TRUE));

	m_bAutoSeparator = GetRCBool("PopupAutoSeparator", TRUE);
	m_bShowExtension = GetRCBool("PopupShowExtension", TRUE);

#ifndef _MINI
	// runitem configuration
	RunItem::SetRememberLast(GetRCBool("PopupRunRememberLast", TRUE));
#endif /* _MINI */

	// Header item configuration
	m_bPinnedAlwaysOnTop = GetRCBool("PinnedPopupNotOnTop", FALSE);
	HeaderItem::SetAlwaysOnTop(m_bPinnedAlwaysOnTop);
	HeaderItem::SetDrawCloseButton(GetRCBool("PopupNoCloseButton", FALSE));

	// POPUP HEIGHTS
	m_nSubmenuHeight = GetRCInt("PopupSubmenuHeight", 20);
	m_nTitleHeight = GetRCInt("PopupTitleHeight", m_nSubmenuHeight);

	// folder item configuration
	int nOverlapX = GetRCInt("PopupOverlapX", 5);
	int nOverlapY = GetRCInt("PopupOverlapY", -m_nSubmenuHeight);
	FolderItem::SetOverlap(nOverlapX, nOverlapY);
	m_bFolderArrow = GetRCBool("PopupNoFolderArrow", FALSE);
	FolderItem::SetDrawArrow(m_bFolderArrow);
	FolderItem::SetDelay(GetRCInt("PopupMenuDelay", 100));

	// datetime configuration
	m_nDateTimeAlignment = GetRCInt("PopupDateTimeAlign", DT_RIGHT);

	// POPUP WIDTHS
	int nMinWidth = GetRCInt("PopupMinWidth", 128);
	PopupMenu::SetMinWidth(nMinWidth);

	if (!GetRCBool("PopupAdaptiveWidth", TRUE))
		PopupMenu::SetMaxWidth(nMinWidth);
	else
		PopupMenu::SetMaxWidth(max(nMinWidth, GetRCInt("PopupMaxWidth", 500)));

#ifndef _MINI
	BmpPainter::SetBlt(GetRCInt("PopupBlt", BMPPAINT_PLAIN));
#endif /* _MINI */
	TaskItem::SetWindowCaption(GetRCBool("PopupTasksWindowCaption", TRUE));

	// more fun with Popup tasks (normalperson)
	TaskItem::SetStayOnSysMenu(GetRCBoolDef("PopupTasksTrackWindowMenu", FALSE));

//	TaskFolder::SetTaskFolderFilterLevel(GetRCInt("PopupTasksFolderFilterLevel", 3));
	TaskFolder::SetTaskFolderTimerInterval(GetRCInt("PopupTasksFolderTimer",2000));
//	m_nTaskFilterLevel = GetRCInt("PopupTasksFilterLevel",12);

//	TaskMenu::SetTaskMenuFilterLevel(GetRCInt("PopupTasksMenuFilterLevel", 3));
	TaskMenu::SetTaskMenuTimerInterval(GetRCInt("PopupTasksMenuTimer",2000));
	TaskFolder::SetSort(GetRCBoolDef("PopupTasksFolderSort",TRUE));
	TaskMenu::SetSort(GetRCBoolDef("PopupTasksMenuSort",TRUE));

	LOGFONT lf;
	SystemParametersInfo(SPI_GETICONTITLELOGFONT, 0, &lf, 0);

	// Load the default font face
	int nFontHeight = GetRCInt("PopupFontHeight", lf.lfHeight);
	int nEntryFontHeight = GetRCInt("PopupEntryFontHeight", nFontHeight);
	int nFolderFontHeight = GetRCInt("PopupFolderFontHeight", nEntryFontHeight);

	GetRCString("PopupFontFace", m_pszFontFace, lf.lfFaceName, MAX_RC_STRING);
	GetRCString("PopupEntryFontFace", m_pszEntryFontFace, m_pszFontFace, MAX_RC_STRING);
	GetRCString("PopupFolderFontFace", m_pszFolderFontFace, m_pszEntryFontFace, MAX_RC_STRING);

	BOOL bTitleBold = GetRCBoolDef("PopupTitleBold", FALSE);

	// title font
	memset(&lf, 0, sizeof(LOGFONT));
	lf.lfHeight = nFontHeight;
	lf.lfWeight = bTitleBold ? FW_BOLD : FW_NORMAL;
	lstrcpyn(lf.lfFaceName, m_pszFontFace, LF_FACESIZE);
	m_hfnTitle = CreateFontIndirect(&lf);

	// entry font
	memset(&lf, 0, sizeof(LOGFONT));
	lf.lfHeight = nEntryFontHeight;
	lstrcpyn(lf.lfFaceName, m_pszEntryFontFace, LF_FACESIZE);
	m_hfnEntry = CreateFontIndirect(&lf);

	// folder font
	memset(&lf, 0, sizeof(LOGFONT));
	lf.lfHeight = nFolderFontHeight;
	lstrcpyn(lf.lfFaceName, m_pszFolderFontFace, LF_FACESIZE);
	m_hfnFolder = CreateFontIndirect(&lf);

	// Set the font
	m_pTitle->SetFont(m_hfnTitle, m_nTitleTextColor);
	m_pEntry->SetFont(m_hfnEntry, m_nEntryTextColor);
	m_pSelEntry->SetFont(m_hfnEntry, m_nSelEntryTextColor);
	m_pFolder->SetFont(m_hfnFolder, m_nFolderTextColor);
	m_pSelFolder->SetFont(m_hfnFolder, m_nSelFolderTextColor);

	// shall bevels be drawn
	MenuItem::DrawBevel(!GetRCBool("PopupNoBevel", TRUE));
	PopupMenu::DrawBevel(!GetRCBool("PopupNoMenuBevel", TRUE));

	// setup colors to be used for 3d details
	MenuItem::SetLight(m_nBevelLightColor);
	MenuItem::SetDark(m_nBevelDarkColor);

	PopupMenu::SetLight(m_nBevelLightColor);
	PopupMenu::SetDark(m_nBevelDarkColor);

	MenuItem::SetIndent(GetRCInt("PopupTextOffset", 6));
	MenuItem::SetRightIndent(GetRCInt("PopupRightIndent", 6));

	char szAlignment[32] =
	    {
	        0
	    };
	GetRCString("PopupEntryAlignment", szAlignment, "", 32);

	if (!lstrcmpi(szAlignment, "center"))
		MenuItem::SetAlignment(DT_CENTER);
	else if (!lstrcmpi(szAlignment, "right"))
		MenuItem::SetAlignment(DT_RIGHT);
	else
		MenuItem::SetAlignment(DT_LEFT);

	GetRCString("PopupTitleAlignment", szAlignment, "", 32);

	if (!lstrcmpi(szAlignment, "center"))
		MenuItem::SetTitleAlignment(DT_CENTER);
	else if (!lstrcmpi(szAlignment, "right"))
		MenuItem::SetTitleAlignment(DT_RIGHT);
	else
		MenuItem::SetTitleAlignment(DT_LEFT);

	GetRCString("HotListName", m_szHotListName, RESSTR(IDS_POPUP_HOTLISTNAME), MAX_RC_STRING);
	GetRCString("PopupTasksNoTitleString", m_szTasksNoTitleString, " [No_Title] ", MAX_RC_STRING);
	PopupMenu::SetScrollSpeed(GetRCInt("PopupScrollSpeed", 1));
	PopupMenu::SetFastKBScrollValue(GetRCInt("PopupFastKBScrollValue", 2));
	int _nMWheelScrollLines = NULL;
	SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, &_nMWheelScrollLines, 0);
	PopupMenu::SetMWheelScroll(GetRCInt("PopupMWheelScrollLines",_nMWheelScrollLines));

//	ReadPopupTasksIgnore();

	// parse and process *Popup config lines in step.rc
	ParseStepRC();


	// assign a default root if none explicitly defined
	if (!m_pRoot)
	{
		if (!m_RootMenus.empty())
			m_pRoot = m_RootMenus[0];
	}

	// register for LS messages
	if (m_pRoot)
	{
		int msgs[] = {
		                 LM_POPUP, LM_HIDEPOPUP, LM_GETREVID, 0
		             };

		SendMessage(hParentWnd, LM_REGISTERMESSAGE, (WPARAM) m_pRoot->GetWindow(), (LPARAM) msgs);
	}

	// create menus for desktops and tasks
	m_pDesktopMenu = new DesktopMenu(RESSTR(IDS_POPUP_DESKTOPS), this);
	m_pDesktopMenu->SetBangName("!PopupDesktops");
	m_RootMenus.push_back(m_pDesktopMenu);

	m_pTaskMenu = new TaskMenu(RESSTR(IDS_POPUP_TASKS), this);
	m_pTaskMenu->SetBangName("!PopupTasks");
	m_pTaskMenu->SetPinnedOnShow(GetRCBool("PopupTasksStartPinned", TRUE) ? TRUE : FALSE);
	m_RootMenus.push_back(m_pTaskMenu);

	SetErrorMode(nErrorMode);
}

/* for dynamic roots (not implemented)

	void PopupMaker::ReadPopupTasksIgnore()
{
	FILE * f = LCOpen(NULL);
	if (!f)
		return;
	unsigned int i;
	char * token[2];
	char buffer[4096];
	numIgnore = 0;

	for (i=0;i<2;++i)
		token[i] = new char[1024];
	while (LCReadNextConfig (f, "*PopupRootTasksIgnore", buffer, sizeof(buffer)))
	{
		int count;
		count = LCTokenize (buffer, token, 2, NULL);
		if (!StripWS(token[1])) continue;
		if (count >= 2)
		{
			if (!ignoreTag)
				ignoreTag = (IgnoreList *)malloc(sizeof(IgnoreList));
			else
				ignoreTag = (IgnoreList *)realloc(ignoreTag, (numIgnore+1)*sizeof(IgnoreList));
			ignoreTag[numIgnore].match = new char[strlen(token[1])+1];
			strcpy(ignoreTag[numIgnore].match, token[1]);
			++numIgnore;
		}
	}
	LCClose(f);
	for (i = 0; i < 2; ++i)
		delete [] token[i];
	return;
} */  

void PopupMaker::ParseStepRC()
{
	FILE *pStepRC = LCOpen(NULL);

	char szLineBuffer[MAX_LINE_LENGTH];
	char szFirst[MAX_PATH] = { 0 };
	char szSecond[MAX_PATH] = { 0 };
	char szThird[MAX_PATH] = { 0 };
	char szFourth[MAX_PATH] = { 0 };

	char *rgszTokens[] = {
	                         szFirst, szSecond, szThird, szFourth
	                     };

	if (!LCReadNextConfig(pStepRC, "*Popup", szLineBuffer, MAX_LINE_LENGTH))
		return ;

	int nTokens = LCTokenize(szLineBuffer, rgszTokens, 4, NULL);

	if (nTokens >= 2 && (lstrcmpi(szSecond, "!New") == 0 || lstrcmpi(szThird, "!New") == 0))
	{
		// multiple popup mode
		do
		{
			char *pszCaption;
			char *pszBangCommand;

			if (nTokens <= 3)
			{
				pszCaption = m_szHotListName;
				pszBangCommand = szThird;
			}
			else
			{
				pszCaption = szSecond;
				pszBangCommand = szFourth;
			}

			PopupMenu *pNewPopup = new PopupMenu(hInst);
			//AddSeparator(pNewPopup);
			pNewPopup->DeleteMenuItems();

			if (pszBangCommand[0])
				pNewPopup->SetBangName(pszBangCommand);
			else
				m_pRoot = pNewPopup;

			AddHeaderItem(pNewPopup, pszCaption, FALSE); // was true
			ParseFolder(pStepRC, pNewPopup);

			pNewPopup->Validate();
			m_RootMenus.push_back(pNewPopup);

			if (!LCReadNextConfig(pStepRC, "*Popup", szLineBuffer, MAX_LINE_LENGTH))
				break;

			szFirst[0] = 0;
			szSecond[0] = 0;
			szThird[0] = 0;
			szFourth[0] = 0;

			nTokens = LCTokenize(szLineBuffer, rgszTokens, 4, NULL);
		}
		while (nTokens >= 2 && (lstrcmpi(szSecond, "!New") == 0 || lstrcmpi(szThird, "!New") == 0));
	}
	else
	{
		// compatibility mode
		m_pRoot = new PopupMenu(hInst);

		m_pRoot->DeleteMenuItems();

		AddHeaderItem(m_pRoot, m_szHotListName, FALSE); // was TRUE
		ParseLine(pStepRC, m_pRoot, szLineBuffer);
		ParseFolder(pStepRC, m_pRoot);
		AddBottomItem(m_pRoot);

		m_pRoot->Validate();
		m_RootMenus.push_back(m_pRoot);
	}

	LCClose(pStepRC);
}

void PopupMaker::ParseFolder(FILE *pStepRC, PopupMenu *pRoot)
{
	while (TRUE)
	{
		char szLineBuffer[MAX_LINE_LENGTH];

		if (!LCReadNextConfig(pStepRC, "*Popup", szLineBuffer, MAX_LINE_LENGTH))
		{
			break;
		}

		if (!ParseLine(pStepRC, pRoot, szLineBuffer))
		{
			break;
		}
	}
}

BOOL PopupMaker::ParseLine(FILE *pStepRC, PopupMenu *pRoot, char *pszLine)
{

	char token1[4096], token2[4096], token3[4096], extra_text[4096];
	LPSTR tokens[4];

	tokens[0] = token1;
	tokens[1] = token2;
	tokens[2] = token3;
//	tokens[3] = token4;

	Painter* pPainter = NULL;
	MenuItem* pMenuItem = NULL;
	char *pszIcon = NULL;
	char szExplicitIcon[MAX_PATH] = { 0 };
	int nHeight = m_nSubmenuHeight;
	int nTokens = 0;
	token1[0] = token2[0] = token3[0] = extra_text[0] = '\0';

	nTokens = LCTokenize (pszLine, tokens, 3, extra_text);

	if (strnicmp(token2, ".icon=", 6) == 0)
	{
		strcpy(szExplicitIcon, token2 + 6);

		char *rgszTokens[1] = { token3 };
		char szBuffer[4096];

		strcpy(token2, token3);
		strcpy(szBuffer, extra_text);
		LCTokenize(szBuffer, rgszTokens, 1, extra_text);

		if (!szBuffer[0])
			nTokens--;
	}

	// special case: separator
	if (nTokens == 2 && lstrcmpi(token2, "!Separator") == 0)
	{
		AddSeparator(pRoot);
		return TRUE;
	}
/* for dynamic roots (not implemented)
	else if (nTokens == 2 && lstrcmpi(token2, "!EVWM") == 0)
	{
		EnumWindows(EnumWindowsProc, (LPARAM)pRoot);
	} */

	if (nTokens == 3)
	{
#ifndef _MINI
		// *popup "Start Menu" 17 default.bmp "!PopupFolder:C:\WINDOWS\Start Menu"
		if (atoi(token3) > 0)
		{
			char pszBmp[4096] = {'\0'};
			char pszCmd[4096] = {'\0'};
			LPSTR ppszExt[3];
			ppszExt[0] = pszBmp;
			ppszExt[1] = pszCmd;

			nHeight = atoi(token3);

			LCTokenize(extra_text, ppszExt, 2, token1);
			pPainter = new BmpPainter(LoadLSImage(pszBmp, NULL), m_bTransparent);
			m_Painters.push_back(pPainter);
			lstrcpy(token3, pszCmd);
		}
#endif /* _MINI */

		if (!lstrcmpi(token3, "Folder"))
		{
			// Folder
			PopupMenu* pSub = new PopupMenu(hInst);

			AddHeaderItem(pSub, token2, FALSE);
			ParseFolder(pStepRC, pSub);
			pSub->Validate();

			pMenuItem = new FolderItem(pSub, token2);
		}
		else if (memicmp(token3, "!PopupFolder:", lstrlen("!PopupFolder:")) == 0)
		{
			// PopupFolder
			char pszFolder[_MAX_PATH];
			LPSTR pszTemp = token3 + lstrlen("!PopupFolder:");
			ExpandEnvironmentStrings(pszTemp, pszFolder, sizeof(pszFolder));

			//PopupMenu* pSub = LoadFolder(token2, pszFolder, extra_text);
			//pMenuItem = new FolderItem(pSub, token2);
			pMenuItem = new DirectoryFolder(pszFolder, FALSE, this, token2, extra_text);
			pszIcon = "";
		}
		else if (memicmp(token3, "!DynamicFolder:", lstrlen("!DynamicFolder:")) == 0)
		{
			// PopupFolder
			char pszFolder[_MAX_PATH];
			LPSTR pszTemp = token3 + lstrlen("!DynamicFolder:");
			ExpandEnvironmentStrings(pszTemp, pszFolder, sizeof(pszFolder));

			//PopupMenu* pSub = LoadFolder(pszFolder);
			pMenuItem = new DirectoryFolder(pszFolder, TRUE, this, token2, extra_text);
			pszIcon = "";
		}
		else if (memicmp(token3, "!PopupDynamicFolder:", lstrlen("!PopupDynamicFolder:")) == 0)
		{
			// PopupFolder
			char pszFolder[_MAX_PATH];
			LPSTR pszTemp = token3 + lstrlen("!PopupDynamicFolder:");
			ExpandEnvironmentStrings(pszTemp, pszFolder, sizeof(pszFolder));

			//PopupMenu* pSub = LoadFolder(pszFolder);
			pMenuItem = new DirectoryFolder(pszFolder, TRUE, this, token2, extra_text);
			pszIcon = "";
		}
		// if it is a special menu item that wants a bmp background
		else if (token2[0] == '!')
		{
			pMenuItem = MakeBangItem(token2, NULL);
#ifndef _MINI
			// token2 == token3 then there is the thing with lscp bugs..
			if (pMenuItem != NULL && lstrcmpi(token2, token3) != 0)
			{
				pPainter = new BmpPainter(LoadLSImage(token3, NULL), m_bTransparent);
				m_Painters.push_back(pPainter);
			}
#endif /* _MINI */
		}
		else
		{
			pMenuItem = MakeBangItem(token3, token2);
			if (pMenuItem == NULL)
			{
				pMenuItem = new CommandItem(token3, extra_text, token2);
				pszIcon = token3[0] == '!' ? NULL : token3;
			}
		}

		if (m_bIcons && pMenuItem)
		{
			HICON hIcon = NULL;

			if (szExplicitIcon[0])
			{
				if (m_bPathIcons)	hIcon = GetIconFromPathEx(szExplicitIcon, m_nIconSize, m_nIconSize);
				// hIcon = LoadLSIcon(szExplicitIcon, NULL);
			}
			else if (!pMenuItem->HasIcon())
			{
				if (pszIcon)
					if (m_bPathIcons)	hIcon = GetIconFromPath(pszIcon);
				else if (m_pszDefaultIcon[0])
					if (m_bPathIcons)	hIcon = GetIconFromPathEx(szExplicitIcon, m_nIconSize, m_nIconSize);
				// hIcon = LoadLSIcon(m_pszDefaultIcon, NULL);
			}

			if (hIcon)
				pMenuItem->SetIcon(hIcon);
		}

		if (pMenuItem != NULL)
		{
			if (pPainter)
			{
				pMenuItem->SetActivePainter(pPainter);
				pMenuItem->SetPainter(pPainter);
			}
			else
			{
				pMenuItem->SetActivePainter(pMenuItem->IsLeaf() ? m_pSelEntry : m_pSelFolder);
				pMenuItem->SetPainter(pMenuItem->IsLeaf() ? m_pEntry : m_pFolder);
			}

			pMenuItem->SetHeight(nHeight);
			pRoot->AddMenuItem(pMenuItem);
		}
	}
	// *popup !DateTime
	else if (nTokens == 2)
	{
		if (lstrcmpi(token2, "~Folder") == 0 || lstrcmpi(token2, "~New") == 0)
		{
			AddBottomItem(pRoot);
			return FALSE;
		}
		else
		{
			pMenuItem = MakeBangItem(token2, NULL);
		}

		if (pMenuItem != NULL)
		{
			pMenuItem->SetActivePainter(pMenuItem->IsLeaf() ? m_pSelEntry : m_pSelFolder);
			pMenuItem->SetPainter(pMenuItem->IsLeaf() ? m_pEntry : m_pFolder);
			pMenuItem->SetHeight(m_nSubmenuHeight);

			if (m_bIcons)
			{
				HICON hIcon = NULL;

				if (szExplicitIcon[0])
					hIcon = LoadLSIcon(szExplicitIcon, NULL);
				else if (m_pszDefaultIcon[0])
					hIcon = LoadLSIcon(m_pszDefaultIcon, NULL);

				if (hIcon)
					pMenuItem->SetIcon(hIcon);
			}

			pRoot->AddMenuItem(pMenuItem);
		}
	}

	return TRUE;
}

Painter* PopupMaker::MakePainter(LPSTR pszImage, DWORD dwColor, DWORD dwGradColor, BOOL bGrad, BOOL bSwap)
{
	Painter* pNewPainter = NULL;
	//DWORD dwLeft = dwColor;
	//DWORD dwRight;

#ifdef _MINI
	pNewPainter = new SolidPainter(dwColor);
#else
	if ( lstrlen(pszImage) <= 0 && !m_bShowGradients)
		pNewPainter = new SolidPainter(dwColor);
  else	if (lstrlen(pszImage) > 0)
	{
		pNewPainter = new BmpPainter(LoadLSImage(pszImage, NULL), m_bTransparent);
	}

	else if ( ( dwGradColor != 0x00ff00ff) && m_bShowGradients)
	{
		DWORD dwLeft = dwColor;
		DWORD dwRight = dwGradColor;
		if (bSwap)
			swap(dwLeft, dwRight);
		pNewPainter = new GradientPainter(dwLeft, dwRight);
	}
	else// if (bGrad && m_bShowGradients)
	{
		DWORD dwLeft = dwColor;
		DWORD dwRight = 0;
		if (bSwap)
			swap(dwLeft, dwRight);
		pNewPainter = new GradientPainter(dwLeft, dwRight);
	}
#endif /* _MINI */
	return pNewPainter;
}

MenuItem* PopupMaker::MakeBangItem(LPSTR pszItem, LPSTR pszTitle)
{
	MenuItem* pMenuItem = NULL;
	if (lstrcmpi(pszItem, "!DateTime") == 0)
		pMenuItem = new TimeItem(m_nDateTimeAlignment, pszTitle == NULL ? "%c" : pszTitle);
#ifndef _MINI
	else if (lstrcmpi(pszItem, "!RemoteAmp") == 0)
		pMenuItem = new RemoteAmpItem();
	else if (lstrcmpi(pszItem, "!PopupImage") == 0)
		pMenuItem = new SeparatorItem(FALSE);
	else if (lstrcmpi(pszItem, "!PopupRun") == 0)
		pMenuItem = new RunItem(pszTitle == NULL ? RESSTR(IDS_POPUP_RUN) : pszTitle);
#endif /* _MINI */
	else if (lstrcmpi(pszItem, "!PopupMyComputer") == 0)
		pMenuItem = new ShellFolder(this, pszTitle, CSIDL_DRIVES);
	else if (lstrcmpi(pszItem, "!PopupNetwork") == 0)
		pMenuItem = new ShellFolder(this, pszTitle, CSIDL_NETWORK);
	else if (lstrcmpi(pszItem, "!PopupPrinters") == 0)
		pMenuItem = new ShellFolder(this, pszTitle, CSIDL_PRINTERS);
	else if (lstrcmpi(pszItem, "!PopupRecentDocuments") == 0)
		pMenuItem = new ShellFolder(this, pszTitle, CSIDL_RECENT);
	else if (lstrcmpi(pszItem, "!PopupRecycleBin") == 0)
		pMenuItem = new ShellFolder(this, pszTitle, CSIDL_BITBUCKET);
	else if (lstrcmpi(pszItem, "!PopupDesktops") == 0)
		pMenuItem = new FolderItem(new DesktopMenu(pszTitle ? pszTitle : RESSTR(IDS_POPUP_DESKTOPS), this), pszTitle ? pszTitle : RESSTR(IDS_POPUP_DESKTOPS));
	else if (strnicmp(pszItem, "!PopupTasks:", strlen("!PopupTasks:")) == 0)
		pMenuItem = new TaskFolder(this, pszTitle == NULL ? RESSTR(IDS_POPUP_TASKS) : pszTitle, atoi( pszItem + strlen("!PopupTasks:") ) );
	else if (lstrcmpi(pszItem, "!PopupTasks") == 0) 
		pMenuItem = new TaskFolder(this, pszTitle == NULL ? RESSTR(IDS_POPUP_TASKS) : pszTitle);
	// else if(lstrcmpi(pszItem, "!Separator") == 0)
	//	pMenuItem = new SeparatorItem(m_pSeparator ? FALSE : TRUE);
	else if (lstrcmpi(pszItem, "!PopupControlPanel") == 0)
		pMenuItem = new ShellFolder(this, pszTitle, CSIDL_CONTROLS);
	return pMenuItem;
}
MenuItem *PopupMaker::AddSeparator(PopupMenu *pMenu)
{
	MenuItem *pMenuItem = new SeparatorItem(m_pSeparator ? FALSE : TRUE);

	if (m_pSeparator)
	{
		pMenuItem->SetActivePainter(m_pSeparator);
		pMenuItem->SetPainter(m_pSeparator);
		pMenuItem->SetHeight(m_nSeparatorHeight);
	}
	else
	{
		pMenuItem->SetActivePainter(m_pSelEntry);
		pMenuItem->SetPainter(m_pEntry);
		pMenuItem->SetHeight(4);
	}

	pMenu->AddMenuItem(pMenuItem);
	return pMenuItem;
}

/* not stable when put into TaskMenu...
MenuItem *PopupMaker::AddTaskFolder(PopupMenu *pMenu)
{
	MenuItem *pMenuItem= new TaskFolder(this, RESSTR(IDS_POPUP_TASKS) );
	pMenuItem->SetActivePainter(m_pSelFolder);
	pMenuItem->SetPainter(m_pFolder);
	pMenuItem->SetHeight(m_nSubmenuHeight);
	
	pMenu->AddMenuItem(pMenuItem);
	return pMenuItem;
} */

void PopupMaker::AddHeaderItem(PopupMenu* pMenu, LPSTR pszTitle, BOOL bVeto)
{
#ifndef _MINI
	if (m_bHeader || bVeto)
	{
		MenuItem* pHead = new HeaderItem(m_bHeaderText ? pszTitle : NULL);
		pHead->SetActivePainter(m_pTitle);
		pHead->SetPainter(m_pTitle);
		pHead->SetHeight(m_nTitleHeight);
		pMenu->AddMenuItem(pHead);
	}
#endif /* _MINI */
}

void PopupMaker::AddBottomItem(PopupMenu* pMenu)
{
#ifndef _MINI
	if (m_pBottom)
	{
		MenuItem* pMenuItem = new SeparatorItem(FALSE);
		pMenuItem->SetPainter(m_pBottom);
		pMenuItem->SetHeight(m_nBottomHeight);
		pMenu->AddMenuItem(pMenuItem);
	}
#endif /* _MINI */
}

PopupMenu* PopupMaker::LoadFolder(LPSTR pszTitle, LPSTR pszFolder, LPSTR pszFilter)
{
	//For folder merging
	char seps[] = "|";
	char *token;
	LPSTR tempPath = new char[strlen(pszFolder) + 1];
	strcpy(tempPath, pszFolder);

	char pszPattern[_MAX_PATH];
	long handle;
	_finddata_t found;
	HICON hIcon = NULL;
	MenuItem* pItem = NULL;
	BOOL bFolders = FALSE;
	BOOL bFiles = FALSE;

	// create menu objects
	PopupMenu* pMenu = new PopupMenu(hInst);

	AddHeaderItem(pMenu, pszTitle, FALSE);

	token = strtok(tempPath, seps );
	while (token != NULL)
	{

		// make file search pattern of foldername
		//lstrcpy(pszPattern, pszFolder);
		lstrcpy(pszPattern, token);
		if (pszPattern[lstrlen(pszPattern)] != '\\')
			lstrcat(pszPattern, "\\");

		// Pattern matching
		if (!pszFilter)
			lstrcat(pszPattern, "*");
		else
		{
			if (strlen(pszFilter) != 0)
				lstrcat(pszPattern, pszFilter);
			else
				lstrcat(pszPattern, "*");
		}

		// for all files in the folder, create a CommandItem
		// and add them to the current popupmenu
		handle = _findfirst(pszPattern, &found);

		if (handle != -1)
		{
			do
			{
				if (lstrcmp(found.name, ".") == 0 || lstrcmp(found.name, "..") == 0 || (found.attrib & _A_HIDDEN))
					continue;
				// if(!(lstrcmp(found.name, ".") && lstrcmp(found.name, "..")) || (found.attrib & _A_HIDDEN))
				//  continue;
				if (found.attrib & _A_SUBDIR)
				{
					bFolders = TRUE;
					char pszSubFolder[_MAX_PATH];
					lstrcpy(pszSubFolder, token);

					if (pszSubFolder[0] && pszSubFolder[lstrlen(pszSubFolder) - 1] != '\\')
						lstrcat(pszSubFolder, "\\");

					lstrcat(pszSubFolder, found.name);
					pItem = new DirectoryFolder(pszSubFolder, FALSE, this, found.name, pszFilter);

					if (m_bIcons)
						pItem->SetIcon(GetIconFromPath(""));
				}
				else
				{
					char pszTitle[_MAX_PATH];
					char pszCmd[_MAX_PATH];

					bFiles = TRUE;
					// Assemble command
					lstrcpy(pszCmd, token);

					if (pszCmd[0] && pszCmd[lstrlen(pszCmd) - 1] != '\\')
						lstrcat(pszCmd, "\\");

					lstrcat(pszCmd, found.name);

					// Fix title (_tremove the extension like .lnk)
					lstrcpy(pszTitle, found.name);
					int nStrLen = lstrlen(pszTitle);
					if (nStrLen > 4 && m_bShowExtension)
					{
						// if we are not to show the extention, just _tremove
						// .lnk or .exe
						LPSTR pszExt = pszTitle + nStrLen - 4;
						if (lstrcmpi(pszExt, ".lnk") == 0 ||
						        lstrcmpi(pszExt, ".pif") == 0 ||
						        lstrcmpi(pszExt, ".bat") == 0 ||
						        lstrcmpi(pszExt, ".exe") == 0 ||
						        lstrcmpi(pszExt, ".com") == 0)
							pszTitle[nStrLen - 4] = '\0';
					}
#ifndef _MINI
					else
						PathRemoveExtension(pszTitle); /* shlwapi dependent */
#endif /* _MINI */
					pItem = new CommandItem(pszCmd, NULL, pszTitle);

					if (m_bIcons)
					{
						SHFILEINFO shfi;
						HIMAGELIST hImageList;

						hImageList = (HIMAGELIST) SHGetFileInfo(pszCmd, 0,
						                                        &shfi, sizeof(SHFILEINFO), SHGFI_ICON | SHGFI_SMALLICON |
						                                        SHGFI_SYSICONINDEX);

						// pItem->SetIcon(GetIconFromPath(pszCmd));
						pItem->SetIcon(hImageList, shfi.iIcon);
					}
				}

				pItem->SetHeight(m_nSubmenuHeight);
				pItem->SetPainter(pItem->IsLeaf() ? m_pEntry : m_pFolder);
				pItem->SetActivePainter(pItem->IsLeaf() ? m_pSelEntry : m_pSelFolder);

				pMenu->AddMenuItem(pItem);

			}
			while (_findnext(handle, &found) == 0);
		}
		_findclose(handle);
		token = strtok(NULL, seps);
	}
	AddBottomItem(pMenu);

	if (m_bAutoSeparator && bFolders && bFiles)
	{
		MenuItem *pMenuItem = AddSeparator(pMenu);
		pMenuItem->SetSortPriority(5);
	}

	pMenu->Sort();
	pMenu->Validate();

	delete [] tempPath;

	return pMenu;
}
