/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/ 
/*
  Copyright (C) 1998-1999 Johan Redestig
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/****************************************************************************
****************************************************************************/
#include <windows.h>
#include "../lsapi/macros.h"
#include "Re5ources.h"
HINSTANCE g_hResStrInstance = NULL;
char g_szResStrBuffer[MAX_PATH];

int GetModkeys(WPARAM wParam)
{
	int modkey = 0x00000000; // bits in human readable format

	if ((LOWORD(wParam) & (MK_CONTROL|MK_SHIFT)) == (MK_CONTROL|MK_SHIFT))
		modkey = 0x00000011;
	else if ((LOWORD(wParam) & MK_CONTROL) == MK_CONTROL)
		modkey = 0x00000001;
	else if ((LOWORD(wParam) & MK_SHIFT) == MK_SHIFT)
		modkey = 0x00000010;
	else if ((LOWORD(wParam) & MK_RBUTTON) == MK_RBUTTON)
		modkey = 0x00000008;
	else modkey = 0x00000000;

	return modkey;
}

BOOL StripWS(char * buffer)	// from jugg's task.dll
{// strip leading and trailing white space from full String
	UINT length = buffer ? strlen (buffer):0;
	while ( length && isspace(*buffer) )
	{
		strcpy (buffer, buffer + 1);
		--length;
	}
	while ( length && isspace(buffer[length-1]) )
		buffer[--length] = 0;
	if ( length ) return TRUE;
	if ( buffer != NULL )
		*buffer = 0;
	return FALSE;
}

BOOL Get4KeyCombo(int vKey0, int vKey1, int vKey2, int vKey3)
{
	return ( ( GetKeyState(vKey0) & KEYDOWN )
	&& ( GetKeyState(vKey1) & KEYDOWN )
	&& ( GetKeyState(vKey2) & KEYDOWN )
	&& ( GetKeyState(vKey3) & KEYDOWN ) );
}

BOOL Get3KeyCombo(int vKey0, int vKey1, int vKey2)
{
	return ( ( GetKeyState(vKey0) & KEYDOWN )
	&& ( GetKeyState(vKey1) & KEYDOWN )
	&& ( GetKeyState(vKey2) & KEYDOWN ) );
}

BOOL Get2KeyCombo(int vKey0, int vKey1)
{
	return ( ( GetKeyState(vKey0) & KEYDOWN )
	&& ( GetKeyState(vKey1) & KEYDOWN ) );
}

HICON GetIconFromPath(const char* pszPath)
{
	SHFILEINFO shFileInfo;
	char pszFileName[MAX_PATH];
	_searchenv(pszPath, "PATH", pszFileName);
	SHGetFileInfo(pszFileName, NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | SHGFI_SMALLICON);

	if (strlen(pszFileName) == 0)
	{
		char pszGuess[MAX_PATH];
		// well... guess that the user has an exe in mind ..
		strcpy(pszGuess, pszPath);
		strcat(pszGuess, ".exe");
		_searchenv(pszGuess, "PATH", pszFileName);
		SHGetFileInfo(pszFileName, NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | SHGFI_SMALLICON);
	} 
	return shFileInfo.hIcon;
}

static int ParseIconLocation(LPSTR szIconFile)
{
	LPSTR pszComma = strrchr(szIconFile, ',');

	if (!pszComma)
		return -1;

	*pszComma = 0;
	return atoi(pszComma + 1);
}

HICON GetIconFromPathEx(LPCSTR pszPath, int cxDesired, int cyDesired)
{
	char szIconFile[MAX_PATH];
	int nIconIndex = -1;
	LPCSTR pszExt;

	strncpy(szIconFile, pszPath, MAX_PATH);
	nIconIndex = ParseIconLocation(szIconFile);
	pszExt = strrchr(szIconFile, '.');

	if (pszExt && stricmp(pszExt, ".ico") == 0)
	{
		return (HICON) LoadImage(NULL, szIconFile, IMAGE_ICON,
		                         cxDesired, cyDesired, LR_LOADFROMFILE);
	}
	else if (nIconIndex >= 0)
	{
		HICON hIcon;

		ExtractIconEx(szIconFile, nIconIndex,
		              (cxDesired <= 24 ? NULL : &hIcon),
		              (cxDesired <= 24 ? &hIcon : NULL), 1);

		return hIcon;
	}
	else
	{
		SHFILEINFO shfi;

		SHGetFileInfo(szIconFile, 0, &shfi, sizeof(SHFILEINFO),
		              SHGFI_ICON | (cxDesired <= 24 ? SHGFI_SMALLICON : 0));

		return shfi.hIcon;
	}
}

HICON GetIconFromWindow(HWND hWnd, BOOL bBigIcon)
{
	int tmp = GetRCInt("PopupGetIconTimeout",1);

	return GetIconFromWindowWithTimeout(hWnd, bBigIcon, tmp);
}

HICON GetIconFromWindowWithTimeout(HWND hWnd, BOOL bBigIcon, int& GetIconTimeout)
{
	HICON hIcon = NULL;

	if ( bBigIcon )
	{
		SendMessageTimeout(hWnd, WM_GETICON, ICON_BIG, 0, SMTO_ABORTIFHUNG, GetIconTimeout, (LPDWORD) &hIcon);
		if (!hIcon)
			hIcon = (HICON) GetClassLong(hWnd, GCL_HICON);
		if (!hIcon)
			SendMessageTimeout(hWnd, WM_QUERYDRAGICON, 0, 0, SMTO_ABORTIFHUNG, GetIconTimeout, (LPDWORD) &hIcon);
		if (!hIcon)
			SendMessageTimeout(hWnd, WM_GETICON, ICON_SMALL, 0, SMTO_ABORTIFHUNG, GetIconTimeout, (LPDWORD) &hIcon);
		if (!hIcon)
			hIcon = (HICON) GetClassLong(hWnd, GCL_HICONSM);
	}
	else
	{
		SendMessageTimeout(hWnd, WM_GETICON, ICON_SMALL, 0, SMTO_ABORTIFHUNG, GetIconTimeout, (LPDWORD) &hIcon);
		if (!hIcon)
			hIcon = (HICON) GetClassLong(hWnd, GCL_HICONSM);
		if (!hIcon)
			SendMessageTimeout(hWnd, WM_QUERYDRAGICON, 0, 0, SMTO_ABORTIFHUNG, GetIconTimeout, (LPDWORD) &hIcon);
		if (!hIcon)
			SendMessageTimeout(hWnd, WM_GETICON, ICON_BIG, 0, SMTO_ABORTIFHUNG, GetIconTimeout, (LPDWORD) &hIcon);
		if (!hIcon)
			hIcon = (HICON) GetClassLong(hWnd, GCL_HICON);
	}
	return hIcon;
}

BOOL IsAppWindow(HWND hWnd)
{
	BOOL bResult = TRUE;
	LONG nStyle;
	HWND hOwner;
	HWND hParent;

	// if it is a WS_CHILD or not WS_VISIBLE, fail it
	nStyle = GetWindowLong(hWnd, GWL_STYLE);
	if ((nStyle & WS_CHILD) || !(nStyle & WS_VISIBLE))
		bResult = FALSE;

	// if the window is a WS_EX_TOOLWINDOW fail it
	nStyle = GetWindowLong(hWnd, GWL_EXSTYLE);
	if (nStyle & WS_EX_TOOLWINDOW)
		bResult = FALSE;

	// If this has an owner, then only accept this window
	// if the partent is not accepted
	hOwner = GetWindow(hWnd, GW_OWNER);
	if (hOwner && bResult)
		bResult = !IsAppWindow(hOwner);

	// if it has a parent, fail
	hParent = GetParent(hWnd);
	if (hParent)
		bResult = FALSE;

	if (GetWindowLong(hWnd, GWL_USERDATA) == 0x49474541)
		bResult = FALSE;

	return bResult;
}

BOOL IsAppWindowWithFilter(HWND hWnd, int filterlevel)
{
	BOOL bResult = TRUE;
	LONG nStyle;
	HWND hOwner;
	HWND hParent;
	/* filter level definitions
		 0 - no filtering at all, titleless windows are given the string " [No Title] "
		 1 - only windows with a title	( EVWM "All Windows" )
		 2 - visible windows, child or not... doesn't matter (EVWM main task list style)
		 3 - original style, only true app windows shown
		12 - Visible windows on desktop, title or not (EVWM main task list while showing windows w/o title)
	*/

	if ( filterlevel > 0) 
	{
		if ( ( filterlevel == 1 ) && ( filterlevel == filterlevel % 10 ) )
		{
			char title[1024];
			if ( GetWindowText(hWnd,title,sizeof(title)) == 0 )
				bResult = FALSE;
		}
		else if ( ( filterlevel > 1)  ) //&& ( filterlevel == filterlevel % 10 ) )
		{
			// if it is a WS_CHILD or not WS_VISIBLE, fail it
			nStyle = GetWindowLong(hWnd, GWL_STYLE);
			if (!(nStyle & WS_VISIBLE))		bResult = FALSE;
			if (nStyle & WS_CHILD) 				bResult = FALSE;
			if (GetWindowLong(hWnd, GWL_USERDATA) == 0x49474541)	bResult = FALSE;

			if ( ( filterlevel > 2 ) && ( filterlevel == filterlevel % 10 ) )
			{
				// If this has an owner, then only accept this window
				// if the partent is not accepted
				hOwner = GetWindow(hWnd, GW_OWNER);
				if (hOwner && bResult)		bResult = !IsAppWindow(hOwner);

				//if (nStyle & WS_CHILD) 	bResult = FALSE;
				// if the window is a WS_EX_TOOLWINDOW fail it
				nStyle = GetWindowLong(hWnd, GWL_EXSTYLE);
				if (nStyle & WS_EX_TOOLWINDOW)		bResult = FALSE;

				// if it has a parent, fail
				hParent = GetParent(hWnd);
				if (hParent)		bResult = FALSE;
			}
		}
	}
	return bResult;
}

int GetInetTime(long now)
{
	// and one hour to get GMT+1 (swizz based time :)
	long nDawn = now + 3600;

	// get number of seconds since midnight
	long nSeconds = nDawn % 86400;

	// convert the number of seconds to a number 0..1000
	int nInetTime = (int)(nSeconds / 86.4);

	// account for possible overlapp
	if (nInetTime >= 1000)
		nInetTime = 0;
	return nInetTime;
}

void SnappWindow(WINDOWPOS* pwPos, int nDist, BOOL bUseScreenSize)
{
	RECT workArea;
	int nTemp;

	if (bUseScreenSize)
	{
		ZeroMemory(&workArea, sizeof(workArea));
		workArea.right = SCREEN_WIDTH;
		workArea.bottom = SCREEN_HEIGHT;
	}
	else
	{
		SystemParametersInfo(SPI_GETWORKAREA, 0, &workArea, 0);
	}

	// left edge
	nTemp = pwPos->x;
	if (nTemp < nDist && nTemp > -nDist)
		pwPos->x = 0;

	// top edge
	nTemp = pwPos->y;
	if (nTemp < nDist && nTemp > -nDist)
		pwPos->y = 0;

	// bottom edge
	nTemp = workArea.bottom - (pwPos->y + pwPos->cy);
	if (nTemp < nDist && nTemp > -nDist)
		pwPos->y = workArea.bottom - pwPos->cy;

	// right edge
	nTemp = workArea.right - (pwPos->x + pwPos->cx);
	if (nTemp < nDist && nTemp > -nDist)
		pwPos->x = workArea.right - pwPos->cx;
}


// tile an image horizontally and vertically
void TileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD dwROP)
{
	for (int y = 0; y < cyDest; y += cySrc)
	{
		for (int x = 0; x < cxDest; x += cxSrc)
		{
			BitBlt(hdcDest,
			       xDest + x,
			       yDest + y,
			       min(cxDest - x, cxSrc),
			       min(cyDest - y, cySrc),
			       hdcSrc,
			       xSrc,
			       ySrc,
			       dwROP);
		}
	}
}
/*
// tile an image horizontally
void HorizontalTileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD dwROP)
{
	for (int x = 0; x < cxDest; x += cxSrc)
	{
		StretchBlt(hdcDest,
		           xDest + x,
		           yDest,
		           min(cxDest - x, cxSrc),
		           cyDest,
		           hdcSrc,
		           xSrc,
		           ySrc,
		           min(cxDest - x, cxSrc),
		           cySrc,
		           dwROP);
	}
}

// tile an image vertically
void VerticalTileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD dwROP)
{
	for (int y = 0; y < cyDest; y += cySrc)
	{
		StretchBlt(hdcDest,
		           xDest,
		           yDest + y,
		           cxDest,
		           min(cyDest - y, cySrc),
		           hdcSrc,
		           xSrc,
		           ySrc,
		           cxSrc,
		           min(cyDest - y, cySrc),
		           dwROP);
	}
}

// multi-blt!
void MultiBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, int cxLeft, int cyTop, int cxRight, int cyBottom, int nMode, DWORD dwROP)
{
	if (cyTop > 0)
	{
		if (cxLeft > 0)
		{
			BitBlt(hdcDest,
			       xDest,
			       yDest,
			       cxLeft,
			       cyTop,
			       hdcSrc,
			       xSrc,
			       ySrc,
			       dwROP);
		}

		if (nMode == MULTIBLT_STRETCH || nMode == MULTIBLT_TILEVERTICAL)
		{
			StretchBlt(hdcDest,
			           xDest + cxLeft,
			           yDest,
			           cxDest - cxLeft - cxRight,
			           cyTop,
			           hdcSrc,
			           xSrc + cxLeft,
			           ySrc,
			           cxSrc - cxLeft - cxRight,
			           cyTop,
			           dwROP);
		}
		else
		{
			HorizontalTileBlt(hdcDest,
			                  xDest + cxLeft,
			                  yDest,
			                  cxDest - cxLeft - cxRight,
			                  cyTop,
			                  hdcSrc,
			                  xSrc + cxLeft,
			                  ySrc,
			                  cxSrc - cxLeft - cxRight,
			                  cyTop,
			                  dwROP);
		}

		if (cxRight > 0)
		{
			BitBlt(hdcDest,
			       cxDest - cxRight,
			       yDest,
			       cxRight,
			       cyTop,
			       hdcSrc,
			       cxSrc - cxRight,
			       ySrc,
			       dwROP);
		}
	}

	if (cxLeft > 0)
	{
		if (nMode == MULTIBLT_STRETCH || nMode == MULTIBLT_TILEHORIZONTAL)
		{
			StretchBlt(hdcDest,
			           xDest,
			           yDest + cyTop,
			           cxLeft,
			           cyDest - cyTop - cyBottom,
			           hdcSrc,
			           xSrc,
			           ySrc + cyTop,
			           cxLeft,
			           cySrc - cyTop - cyBottom,
			           dwROP);
		}
		else
		{
			VerticalTileBlt(hdcDest,
			                xDest,
			                yDest + cyTop,
			                cxLeft,
			                cyDest - cyTop - cyBottom,
			                hdcSrc,
			                xSrc,
			                ySrc + cyTop,
			                cxLeft,
			                cySrc - cyTop - cyBottom,
			                dwROP);
		}
	}

	if (nMode == MULTIBLT_STRETCH)
	{
		StretchBlt(hdcDest,
		           xDest + cxLeft,
		           yDest + cyTop,
		           cxDest - cxLeft - cxRight,
		           cyDest - cyTop - cyBottom,
		           hdcSrc,
		           xSrc + cxLeft,
		           ySrc + cyTop,
		           cxSrc - cxLeft - cxRight,
		           cySrc - cyTop - cyBottom,
		           dwROP);
	}
	else if (nMode == MULTIBLT_TILE)
	{
		TileBlt(hdcDest,
		        xDest + cxLeft,
		        yDest + cyTop,
		        cxDest - cxLeft - cxRight,
		        cyDest - cyTop - cyBottom,
		        hdcSrc,
		        xSrc + cxLeft,
		        ySrc + cyTop,
		        cxSrc - cxLeft - cxRight,
		        cySrc - cyTop - cyBottom,
		        dwROP);
	}
	else if (nMode == MULTIBLT_TILEHORIZONTAL)
	{
		HorizontalTileBlt(hdcDest,
		                  xDest + cxLeft,
		                  yDest + cyTop,
		                  cxDest - cxLeft - cxRight,
		                  cyDest - cyTop - cyBottom,
		                  hdcSrc,
		                  xSrc + cxLeft,
		                  ySrc + cyTop,
		                  cxSrc - cxLeft - cxRight,
		                  cySrc - cyTop - cyBottom,
		                  dwROP);
	}
	else if (nMode == MULTIBLT_TILEVERTICAL)
	{
		VerticalTileBlt(hdcDest,
		                xDest + cxLeft,
		                yDest + cyTop,
		                cxDest - cxLeft - cxRight,
		                cyDest - cyTop - cyBottom,
		                hdcSrc,
		                xSrc + cxLeft,
		                ySrc + cyTop,
		                cxSrc - cxLeft - cxRight,
		                cySrc - cyTop - cyBottom,
		                dwROP);
	}

	if (cxRight > 0)
	{
		if (nMode == MULTIBLT_STRETCH || nMode == MULTIBLT_TILEHORIZONTAL)
		{
			StretchBlt(hdcDest,
			           cxDest - cxRight,
			           yDest + cyTop,
			           cxRight,
			           cyDest - cyTop - cyBottom,
			           hdcSrc,
			           cxSrc - cxRight,
			           ySrc + cyTop,
			           cxRight,
			           cySrc - cyTop - cyBottom,
			           dwROP);
		}
		else
		{
			VerticalTileBlt(hdcDest,
			                cxDest - cxRight,
			                yDest + cyTop,
			                cxRight,
			                cyDest - cyTop - cyBottom,
			                hdcSrc,
			                cxSrc - cxRight,
			                ySrc + cyTop,
			                cxRight,
			                cySrc - cyTop - cyBottom,
			                dwROP);
		}
	}

	if (cyBottom > 0)
	{
		if (cxLeft > 0)
		{
			BitBlt(hdcDest,
			       xDest,
			       cyDest - cyBottom,
			       cxLeft,
			       cyBottom,
			       hdcSrc,
			       xSrc,
			       cySrc - cyBottom,
			       dwROP);
		}

		if (nMode == MULTIBLT_STRETCH || nMode == MULTIBLT_TILEVERTICAL)
		{
			StretchBlt(hdcDest,
			           xDest + cxLeft,
			           cyDest - cyBottom,
			           cxDest - cxLeft - cxRight,
			           cyBottom,
			           hdcSrc,
			           xSrc + cxLeft,
			           cySrc - cyBottom,
			           cxSrc - cxLeft - cxRight,
			           cyBottom,
			           dwROP);
		}
		else
		{
			HorizontalTileBlt(hdcDest,
			                  xDest + cxLeft,
			                  cyDest - cyBottom,
			                  cxDest - cxLeft - cxRight,
			                  cyBottom,
			                  hdcSrc,
			                  xSrc + cxLeft,
			                  cySrc - cyBottom,
			                  cxSrc - cxLeft - cxRight,
			                  cyBottom,
			                  dwROP);
		}

		if (cxRight > 0)
		{
			BitBlt(hdcDest,
			       cxDest - cxRight,
			       cyDest - cyBottom,
			       cxRight,
			       cyBottom,
			       hdcSrc,
			       cxSrc - cxRight,
			       cySrc - cyBottom,
			       dwROP);
		}
	}
} // */

void TrimLeft( char *toTrim )
{
	char *trimmed = toTrim;

	// skip past spaces
	while ( *toTrim && isspace( *toTrim ) )
		++toTrim;

	// copy the rest of the string over
	while ( *toTrim )
		*++trimmed = *++toTrim;

	// null-terminate it
	*trimmed = 0;
}

LPSTR LoadStringInplace( HINSTANCE hInstance, UINT nID, LPSTR pszBuffer, UINT nLength )
{
	LoadString( hInstance, nID, pszBuffer, nLength );
	return pszBuffer;
}

